#!/bin/bash -e 
#
# Yet Another Pool Compatible Anon Rsync Partial Debian Mirror Script
# absurd-debmirror
#
# Special dependencies: rsync.deb, procmail.deb (lockfile)
#
# (c) 1999, 2000, 2001 Stephan A Suerken <absurd@debian.org>, GPL.
# Copyright 1999, 2000 by Joey Hess <joeyh@debian.org> (gathering filelist)
# Originated from ancient script "anonftpsync" somewhere from www.debian.org/mirror/
#
# Updates: http://www.Fh-Worms.DE./~inf222/code/script/debian/www/debian.en.html
#
# TODO: Add this to CVS!
#
# What is this?
# =============
# This is a partial rsync mirror script for the Debian Archive, for use during (and
# hopefully after) the transition to package pools. Can be used on different rsync
# archives/dists/type via arguments, takes some special care to sync worthwile
# dist/type dependent data (proposed updates, disk images,...).
#
# How to use
# ==========
# 1. Change the two configurational lines below.
# 2. Install the script as <mirror>.
# 3. Run "<mirror>", read the usage help.
#
#    Example:
#
#    <mirror> "/home/ftp/pub"         "ftp.de.debian.org"                             "debian-non-US" \
#             "potato woody sid"      "non-US/main non-US/contrib non-US/non-free"    "source i386"
#
# would mirror the debian-non-US (rsync) archive to the local dir /home/ftp/pub/debian-non-US, syncing
# source and i386 from all sections of potato, woody and sid.
# 
# The script is designed idempotent, i.e., you could run it with a different selection on the same
# archive; the clean up will only delete packages files not in any existing Packages.gz or Sources.gz.
#
# Known Problems
# ==============
# o File deletion. See Changes.
# o There might be a problem with newer rsync servers (need to verify).
# o See TODO-labels in script ;).
# o Generally, the script is rather a hog...
#
# Changes
# =======
# 2001.02.06.1: o Now getting proposed-updates, too, via packages files. Fixes files
#                 in proposed updates that have moved to pool not to be mirrored.
#                 Feature not thoroughly tested.
# 2001.01.26.1: o Using "--copy-links" instead of "--links" as default when syncing packages
#                 ==> Safer. Fixes some cases of missing packages (when filename in Packages file
#                 is actually a symlink), might retrieve some files more than once.
#               o File deletion per default only in ^pool/. 
#                 ==> Safer. Fixes wrong deletion of files which are not "correctly" listed in
#                 Packages files (e.g. files listed as "dist/stable", packages in proposed-updates),
#                 but obsolete packages elsewhere than ^pool/ will stick.
# 2001.01.07.1: Improved package file creation:
#                 - Own function for list generation 
#                 - Duplicates are removed from packages file (makes them ~ dists*2 smaller)
#                 - Not running mkdir on all dirs => making rsync comp. dir skeleton is about
#                   10*faster for my setup)
# 2000.12.28.3: First "release". Was called "2000.28.12.3".
#
#
#----------------------------------------------------------------------
# Configuration
#
SYNC_RUNUSER=absurd
DISK_IMAGES_EXCLUDES="--exclude **images-1.20/ --exclude **images-1.88/"

#----------------------------------------------------------------------
# There should be no need to touch anything below
#

# Sync won't start if this file is present. Will be removed (trapped) on script
# exit, but that is not neccesarily fool proof.
SYNC_LOCKFILE=".sync_LOCK"
# If sync succedded, this file will be touched plus the mirror host name.
SYNC_TRACEFILE=".sync_trace_of_"

#
# Collect arguments
#
# "/home/ftp/pub"
SYNC_DIR=$1

# "ftp.de.debian.org"
SYNC_HOST=$2

# "debian" "debian-non-US"
SYNC_ARCHIVE=$3

# "slink potato woody"
SYNC_DISTS=$4

# "main contrib non-free"
SYNC_SECTIONS=$5

# "source i386 m68k"
SYNC_TYPES=$6


SYNC_DEST=${SYNC_DIR}/${SYNC_ARCHIVE}


#
# Arguments to rsync
#
# For tests: --partial		Do NOT remove (and DO continue later) partially
#				files.
#            --timeout=n	Leave after a n seconds of no traffic.
#            --verbose		Increase verbosity level.
#            --dry-run		Test run
#            --delete		This should normally be given. Is dangerous in case
#                               the rsyncd file list is bogus (?!).

#
# rsync usage
#
# o "mirroring" configuration (sort of)
# o Choose other options explicitly (especially: recursion, symlink handling)
#RSYNC="rsync --perms --times --links --partial --verbose"
RSYNC="rsync --perms --times --partial --timeout=300 --verbose"


#
# Part ???: Closeup, misc
#

# I think this gives you sort of orientation in the output
# Many will think this is ugly, but it's actually ascii art.
function echo_header () # 1: level  2: text
{
  echo
  echo "$2"
  case $1 in
       0) echo "**********************************************************************"
          ;;
       1) echo "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
          ;;
       2) echo "======================================================================"
          ;;
       3) echo "----------------------------------------------------------------------"
          ;;
       *) ;;
  esac
}

# This removes the lock file
function remlock
{
    rm -f ${SYNC_DEST}/${SYNC_LOCKFILE}
    echo "Lockfile removed: ${SYNC_LOCKFILE}"
}

# standard message unless specified otherwise after successful run at the end of script
SYNC_CLOSEUP_MESSAGE="Mirror Run Interrupted Or Bogus."

function sync_closeup
{
    echo_header 1 "Mirror Closeup"
    remlock

    echo_header 2 "Removing Dangling Symlinks"
    symlinks -r -d ${SYNC_DEST}

    echo_header 2 "Disk Usage Statistic (`date`)"
    # Old style dists
    for dist in ${SYNC_DISTS} ; do                           # potato woody ...
        du --summarize --human-readable dists/${dist}*
    done
    # All other packages already in pool
    du --summarize --human-readable pool

    echo
    echo_header 0 "$SYNC_CLOSEUP_MESSAGE"
}

# Sorts a given file and uniqifies it
function sync_sort_filelist ()
{
    local file=$1
    local tempfile=`tempfile`

    echo -n "Sorting and removing duplicates in ${file}..."
    sort ${file} | uniq > ${tempfile}
    cp ${tempfile} ${file}
    rm -f ${tempfile}
    echo "done."
}

# Generates a file list from a list of packages files
function sync_gen_filelist ()
{
    local file=$1
    local desc=$2
    local source_pf=$3
    local binary_pf=$4

# debug
    local date_pre=`date +%s`

    echo
    echo "Filelist generation: ${desc}"
    echo
    echo "  Source: ${source_pf}"
    echo "  Binary: ${binary_pf}"
    echo
    echo -n "Generating \"${file}\"..."

    # Make sure the file exists with size 0
    cat /dev/null >${file}
    for filename in `echo "${binary_pf}" | xargs -r zgrep -i ^Filename: | cut -d ' ' -f 2` \
	            `echo "${source_pf}" | xargs -r zcat | \
		     awk '/^Directory:/ {D=$2} /Files:/,/^$/ { if ($1 != "Files:" && $0 != "") print D "/" $3; }'`
         do
         # Note: Using non-bash-builtins here will slow down performance significantly (about 10 times per command)
         echo ${filename} >> ${file}
    done
    echo "done."

    sync_sort_filelist ${file}

    # debug
    echo "The generation took $((`date +%s` - ${date_pre})) seconds."
}


# these four globals are assembled here, used in phase III
BINARY_PACKAGES_FILES=""
SOURCE_PACKAGES_FILES=""
BINARY_PACKAGES_FILES_PROPUP=""
SOURCE_PACKAGES_FILES_PROPUP=""

function sync_packages_file ()
{
   local path=$1
   local type=$2
   local propup=$3
   local packages_file=""
   local release_file=""

   if [ -n "${propup}" ] ;
      then if [ "${type}" = "source" ] ;
              then packages_file=${path}/Sources.gz
                   release_file=${path}/Release
                   SOURCE_PACKAGES_FILES_PROPUP="${SOURCE_PACKAGES_FILES_PROPUP} ${packages_file}"
              else packages_file=${path}/Packages.gz
                   release_file=${path}/Release 
                   BINARY_PACKAGES_FILES_PROPUP="${BINARY_PACKAGES_FILES_PROPUP} ${packages_file}"
           fi
      else if [ "${type}" = "source" ] ;
              then packages_file=${path}/${type}/Sources.gz
                   release_file=${path}/${type}/Release 
                   SOURCE_PACKAGES_FILES="${SOURCE_PACKAGES_FILES} ${packages_file}"
              else packages_file=${path}/binary-${type}/Packages.gz
                   release_file=${path}/binary-${type}/Release 
                   BINARY_PACKAGES_FILES="${BINARY_PACKAGES_FILES} ${packages_file}"
           fi
   fi

   # || true needed, as we use it universally (e.g., as well for sid-proposed-updates)   
   mkdir -p `dirname ${packages_file}` || true
   ${RSYNC} --copy-links ${SYNC_HOST}::${SYNC_ARCHIVE}/${packages_file} ${packages_file} || true
   ${RSYNC} --copy-links ${SYNC_HOST}::${SYNC_ARCHIVE}/${release_file}  ${release_file} || true
}


# Use this for hilarious debugging sessions
function sync_debug ()
{
 echo_header 1 "DEBUGGING"
 ############################################

 exit 0
}


#############################################################
#                     Start Script                          #
#############################################################

#
# Check: Usage
#
if [ "$2" == "" ]; 
   then cat <<EOF
Usage 0: `basename $0` <dir> <host> <archive> <dists> <sections> <types>
         <dir>     = Dest directory                    "/home/ftp/pub"
         <host>    = Rsync host                        "ftp.de.debian.org"
         <archive> = Rsync archive, subdir in <dir>    "debian", "debian-non-US"
         <dists>   = List. Distribution codenames.     "potato woody"
                     !The real, non-symlink codenames!
         <sections>= List. Sections.                   "main contrib non-free"
                     There are 6 possible.             "non-US/main non-US/contrib non-US/non-free"
         <types>   = List. Source+Architectures.       "source i386 m68k"
EOF
      exit 1
fi

#
# Check: User ID
# 
if [ "`id --name --user`" != "${SYNC_RUNUSER}" ] ;
   then echo "Error: Must be run as user ${SYNC_RUNUSER}!" >&2
	exit 1
fi

#
# Check: Lockfile
#
if ! lockfile -l 43200 -r 0 "${SYNC_DEST}/${SYNC_LOCKFILE}" ;
   then echo "Error mirroring ${SYNC_DEST}: Lockfile exists!"
        exit 1
fi

#
# Everything's fine so far. Let's try it
#

# We need this minimal skeleton
mkdir -p ${SYNC_DEST}/dists ${SYNC_DEST}/pool

# Now we do everything inside destination dir
cd ${SYNC_DEST} || exit 1

#
# Change sync_debug and uncomment this to "debug"
#
#sync_debug

# 
umask 002

# trap closeup _now_. It's the (textual) counterpart of "Starting New Mirror"
trap "sync_closeup" EXIT

echo_header 0 "Starting New Mirror Run"
echo "Start time : `date`"
echo "Mirror host: $SYNC_HOST"
echo "Archive    : $SYNC_ARCHIVE"
echo "Dists      : $SYNC_DISTS"
echo "Sections   : $SYNC_SECTIONS"
echo "Types      : $SYNC_TYPES"
echo "Dest       : $SYNC_DEST"
echo


echo_header 1 "Phase I: Debian Global Files"
############################################

# Readmes etc
echo_header 2 "Top Level Files (one level)"
${RSYNC} --links --delete \
         --exclude "/ls-lR*" \
         --exclude "/Archive-Update-in-Progress*" \
         "${SYNC_HOST}::${SYNC_ARCHIVE}/*" .

# Ensures all dists dirs/symlinks are up-to date.
echo_header 2 "Top Level Dists Dir (one level)"
${RSYNC} --links "${SYNC_HOST}::${SYNC_ARCHIVE}/dists/*" dists/

echo_header 2 "Top Level Doc Dir (complete)"
${RSYNC} --recursive --links --delete "${SYNC_HOST}::${SYNC_ARCHIVE}/doc/*" doc/ || true

echo_header 2 "Top Level Tools Dir (complete)"
${RSYNC} --recursive --links --delete "${SYNC_HOST}::${SYNC_ARCHIVE}/tools/*" tools/ || true


echo_header 1 "Phase II: Selected Global Files"
###############################################

for dist in ${SYNC_DISTS} ; do                           # potato woody ...
    for type in ${SYNC_TYPES} ; do                       # alpha i386 source ...

        echo_header 2 "Global Files For: ${dist}/${type}"

        echo_header 3 "Contents, Release, Changelog, etc"
        ${RSYNC} --links --delete \
                 --include "Contents-${type}.gz" \
                 --exclude "Contents-*.gz" \
                 --include "*" \
                 "${SYNC_HOST}::${SYNC_ARCHIVE}/dists/${dist}/*" dists/${dist}/

        if [ "${type}" = "source" ] ;
	    then echo_header 3 "Packages File for Proposed Updates"
                 sync_packages_file "dists/${dist}-proposed-updates" "${type}" PROPUP

#                 ${RSYNC} --recursive --links --delete \
#                          --include "*.dsc" \
#                          --include "*.diff.gz" \
#                          --include "*.orig.tar.gz" \
#                          --include "Sources.gz" \
#                          --exclude "*" \
#                          "${SYNC_HOST}::${SYNC_ARCHIVE}/dists/${dist}-proposed-updates/*" \
#                          dists/${dist}-proposed-updates/ || true

            else echo_header 3 "Upgrade Directory"
                 main_path=dists/${dist}/main
                 mkdir -p ${main_path} || true
                 ${RSYNC} --recursive --links --delete \
                          "${SYNC_HOST}::${SYNC_ARCHIVE}/${main_path}/upgrade-${type}/*" \
                          ${main_path}/upgrade-${type}/ || true

                 echo_header 3 "Disk Images"
                 ${RSYNC} --recursive --links --delete \
                          ${DISK_IMAGES_EXCLUDES} \
                          "${SYNC_HOST}::${SYNC_ARCHIVE}/${main_path}/disks-${type}/*" \
                          ${main_path}/disks-${type}/ || true

                 echo_header 3 "Packages File for Proposed Updates"
                 sync_packages_file "dists/${dist}-proposed-updates" "${type}" PROPUP

#                 ${RSYNC} --recursive --links --delete \
#                          --include "*_all.deb" \
#                          --include "*_${type}.deb" \
#                          --include "Packages.gz" \
#                          --exclude "*" \
#                          "${SYNC_HOST}::${SYNC_ARCHIVE}/dists/${dist}-proposed-updates/*" \
#                          dists/${dist}-proposed-updates/ || true
        fi

        echo_header 3 "Packages Files"
        for section in ${SYNC_SECTIONS} ; do              # main contrib non-free ...
            sync_packages_file "dists/${dist}/${section}" "${type}"
	done
    done
done


echo_header 1 "Phase III: Updating Packages"
############################################

# a-0) generate proposed-updates filelist to sync
pfpu=.sync_pfpu
sync_gen_filelist "${pfpu}" "All package filenames in proposed updates" \
                  "${SOURCE_PACKAGES_FILES_PROPUP}" "${BINARY_PACKAGES_FILES_PROPUP}"
pfpu2=.sync_pfpu2
rm -f ${pfpu2}
for type in ${SYNC_TYPES} ; do                       # alpha i386 source ...
    if [ "${type}" = "source" ] ;
	then grep -e ".*\.dsc" -e ".*\.diff\.gz" -e ".*\.orig\.tar\.gz" ${pfpu} >> ${pfpu2}
        else grep -e ".*_all\.deb" -e ".*_${type}\.deb" ${pfpu} >> ${pfpu2}
    fi
done

# a-1) generate "normal" filelist to sync
pf=.sync_pf
sync_gen_filelist "${pf}" "All package filenames to be synced in this run" \
                  "${SOURCE_PACKAGES_FILES}" "${BINARY_PACKAGES_FILES}"

# a) merge the two package lists
cat ${pfpu2} >> ${pf}
sync_sort_filelist ${pf}


# b) rsync compatability
# TODO: fear rsync really needs these dirs to exist. ?? cant rsync do this?
echo
echo "Creating missing directories:"
echo "Note: (PRE-POOL) You might want to check if some of these are actually symlinks."
date_pre=`date +%s`
for f in `cat ${pf}`; do
    # Running mkdir only on non-existing dirs: speeds up things greatly
    [ ! -e ${f} ] && mkdir --verbose --parents `dirname $f`
done
echo "Dirmaking took $((`date +%s` - ${date_pre})) seconds."

# c) now rsync the files in the filelist
# TODO: Using "--links": If filename is actually a symlink its not guaranteed that
# TODO: the actual file is really retrieved.
# TODO: Using "--copy-links": If filename is actually a symlink, you might download
# TODO: the same file several times (but it will work).
# TODO: afaik, there should not be any symlinks in Packages list files. In the future
# TODO: at least.
echo
${RSYNC} --recursive --copy-links \
         --include-from ${pf} \
         --exclude "*" \
         ${SYNC_HOST}::${SYNC_ARCHIVE}/ .

#----------------------------------------------------------------------
echo_header 1 "Phase IV: Clean Up"

# TODO: It only cleans up obsolete Packages files.
# TODO: Can "rsync --recursive --delete ..." be used somehow only to DELETE
#       files not in the rsync hosts complete filelist?

# a) generate should-be-installed list
pf_packages=.sync_pf_packages
sync_gen_filelist "${pf_packages}" \
                  "Filenames from all local Packages.gz/Sources.gz." \
                  "`find . -type f -name Sources.gz`" \
                  "`find . -type f -name Packages.gz`"

# b) generate is-installed-list
echo
echo "Building filelist: Filenames of all locally installed packages files..."
pf_installed=.sync_pf_installed
# Make sure the files are there && empty
rm -f ${pf_installed}
touch ${pf_installed}
# TODO: Eventually, all pakages are actually in pool/, and dists/ can go away here.
# TODO: THEN: In pool/, all files are plain and packages (really?) ==>
# TODO: THEN: Might be ok even to just use all files then?
for file in `find dists/ pool/ -name "*.deb" -or -name "*.dsc" -or -name "*.diff.gz" -or -name "*.orig.tar.gz"` ; do
    echo "${file}" >> ${pf_installed}
done
sync_sort_filelist ${pf_installed}

# c) generate obolete list
echo
echo "Building filelist: Filenames of obsolete packages files..."
pf_obsolete=.sync_pf_obsolete
# Make sure the files are there && empty
rm -f $pf_obsolete
touch $pf_obsolete
# comm  |<Column
#       1: Only in file1
#       2: Only in file2
#       3: In file1 AND in file2
# Arg -n: supresses column n.
comm -2 -3 ${pf_installed} ${pf_packages} > ${pf_obsolete}

# d) remove obsolete files
# CAVEATS: As long as we are still deleting in dists, packages that are not mentioned
#          in any Packages files get deleted. Currently affected afaik:
#  1) proposed-updates (i.e., if the Packages file is not correct)
#  2) Packages mentioned as dists/<symlink>/... in Packages files
#  3) .deb-files in "upgrade" directories
#
echo
echo "Removing obsolete files:"
echo
# Only deleting in pool will cure 1), 2)
cat ${pf_obsolete} | grep "^pool/"         | xargs rm -f -v
# Adding these will make you vulnerable for 2), but still cures 1)
#cat ${pf_obsolete} | grep "^dists/potato/" | xargs rm -f -v
#cat ${pf_obsolete} | grep "^dists/woody/"  | xargs rm -f -v
#cat ${pf_obsolete} | grep "^dists/sid/"    | xargs rm -f -v

# c) clean up clean up
#
# do not delete as long as we are testing
#rm -f ${pf} ${pf_packages} ${pf_installed} ${pf_obsolete}

#----------------------------------------------------------------------
# Well done.
#
SYNC_CLOSEUP_MESSAGE="Mirror Run Successfully Finished."
date -u > "${SYNC_TRACEFILE}${SYNC_HOST}"
exit 0
